use crate::*;

/// define blockchain subcommands
#[derive(Clone, Default, Debug, clap::Parser)]
pub enum Subcommand {
	#[clap(hide = true)]
	Repart {
		// Number of transactions per block to target
		target: u32,
		#[clap(short = 'o', long = "old-repart")]
		// Old/actual repartition
		actual_repart: Option<u32>,
	},
	#[clap(hide = true)]
	SpamRoll { actual_repart: usize },
	/// Get information about runtime
	RuntimeInfo,
	/// Check current block
	#[default]
	CurrentBlock,
	/// Create one block manually (manual sealing)
	#[clap(hide = true)]
	CreateBlock,
}

/// handle blockchain commands
pub async fn handle_command(data: Data, command: Subcommand) -> Result<(), GcliError> {
	let mut data = data.build_client().await?;
	match command {
		Subcommand::Repart {
			target,
			actual_repart,
		} => commands::net_test::repart(&data, target, actual_repart).await?,
		Subcommand::SpamRoll { actual_repart } => {
			commands::net_test::spam_roll(&data, actual_repart).await?
		}
		Subcommand::RuntimeInfo => {
			data = data.fetch_system_properties().await?;
			commands::runtime::runtime_info(data).await;
		}
		Subcommand::CurrentBlock => {
			println!(
				"current block on {}: {}",
				data.cfg.duniter_endpoint,
				data.client()
					.storage()
					.at_latest()
					.await?
					.fetch(&runtime::storage().system().number())
					.await?
					.unwrap()
			);
		}
		Subcommand::CreateBlock => {
			todo!()
			// data.client()
			// 	.backend()
			// 	.call("engine_createBlock", subxt::backend::rpc::rpc_params![true, true]) // create empty block and finalize
			// 	.await?; // FIXME this gives a serialization error
		}
	}
	Ok(())
}

/// get genesis hash
pub async fn fetch_genesis_hash(data: &Data) -> Result<Hash, anyhow::Error> {
	Ok(data
		.client()
		.storage()
		.at_latest()
		.await?
		.fetch(&runtime::storage().system().block_hash(0))
		.await?
		.unwrap())
}
